from abc import ABC, abstractmethod, abstractclassmethod
from typing import Generator, Mapping, Any

import pandas as pd


class ColumnInfo:
    def __init__(
        self,
        database,
        schema,
        table,
        name,
        position,
        dtype,
        nullable: bool = True,
        extra: Mapping[str, Any] = None,
    ):
        self.database = database
        self.table = table
        self.schema = schema
        self.name = name
        self.position = int(position)
        self.nullable = bool(nullable)
        self.dtype = dtype
        self.extra = extra or {}

    def __str__(self) -> str:
        return f"Column: {self.name}, position: {self.position}, nullable: {self.nullable}, dtype: {self.dtype}"

    def __repr__(self) -> str:
        return self.__str__()

    def to_dict(self) -> str:
        return {
            "database": self.database,
            "schema": self.schema,
            "table": self.table,
            "name": self.name,
            "position": self.position,
            "nullable": self.nullable,
            "dtype": self.dtype,
            "extra": self.extra,
        }


class BaseAdapter(ABC):
    def __init__(self, database) -> None:
        self.database = database

    @abstractclassmethod
    def from_dbt_profile(cls, profile: Mapping[str, Any], database=None):
        pass

    @abstractmethod
    def list_datasets(self) -> Generator[str, None, None]:
        pass

    @abstractmethod
    def list_tables(self, dataset: str) -> Generator[str, None, None]:
        pass

    @abstractmethod
    def list_columns(self, dataset: str, table: str) -> Generator[ColumnInfo, None, None]:
        pass

    @abstractmethod
    def execute(self, query, **kwargs) -> Any:
        pass

    @abstractmethod
    def execute_pandas(self, query, **kwargs) -> pd.DataFrame:
        pass



from functools import lru_cache

from google.cloud import bigquery
import pandas as pd
from typing import Generator

from .base import ColumnInfo, BaseAdapter


class BigQueryAdapter(BaseAdapter):
    def __init__(self, database, location="US"):
        self.database = database
        self.location = location
        self._client = None

    @property
    def client(self) -> bigquery.Client:
        if not self._client:
            self._client = bigquery.Client(project=self.database, location=self.location)
        return self._client

    @classmethod
    def from_dbt_profile(cls, profile: dict, database=None):
        database = database or profile["project"]
        return cls(database=database, location=profile.get("location", "US"))

    def execute(self, query, **kwargs):
        return self.client.query(query, **kwargs).result()

    def execute_pandas(self, query, **kwargs) -> pd.DataFrame:
        return self.client.query(query, **kwargs).result().to_dataframe(bqstorage_client=True)

    @lru_cache()
    def get_column_metadata_df(self, dataset) -> pd.DataFrame:
        query = "select *" f" from `{self.database}.{dataset}.INFORMATION_SCHEMA.COLUMNS`;"
        dataframe = self.execute_pandas(query)
        dataframe = dataframe.loc[
            dataframe["table_schema"].apply(lambda x: (x.lower() != "information_schema" and not x.startswith("_")))
            & dataframe["table_name"].apply(lambda x: not x.startswith("_")),
            :,
        ]
        dataframe.sort_values(by=["table_schema", "table_name", "ordinal_position"], inplace=True)
        return dataframe

    def list_datasets(self) -> Generator[str, None, None]:
        datasets = list(self.client.list_datasets(self.database))
        datasets.sort(key=lambda x: x.dataset_id)
        for dataset in datasets:
            yield dataset.dataset_id

    def list_tables(self, dataset: str) -> Generator[str, None, None]:
        meta_df = self.get_column_metadata_df(dataset)
        for schema in meta_df["table_name"].unique():
            yield schema

    def list_columns(self, dataset: str, table: str) -> Generator[ColumnInfo, None, None]:
        meta_df = self.get_column_metadata_df(dataset)
        meta_df = meta_df[meta_df["table_name"] == table]
        rows = meta_df.to_dict(orient="records")
        for row in rows:
            yield ColumnInfo(
                database=self.database,
                schema=dataset,
                table=table,
                name=row["column_name"],
                position=row["ordinal_position"],
                dtype=row["data_type"],
                nullable=True if row["is_nullable"] == "YES" else False,
            )



import os
from concurrent.futures import ThreadPoolExecutor, as_completed

from tqdm import tqdm

from .adapters import get_adapter
from .templater import get_template
from .utils import DEFAULT_DBT_PROFILE_PATH, read_dbt_profile, read_yaml, write_file

template = get_template("base_model.sql")


def read_source_yaml(path):
    source_config = read_yaml(path)

    if "sources" not in source_config:
        raise ValueError("Source YAML malformed. Missing `sources`.")

    return source_config["sources"]


def make_model_sql(*args, **kwargs):
    sql = template.render(*args, **kwargs)
    return sql


def generate_base_from_sources(profile, output_folder, source_path):
    sources = read_source_yaml(source_path)

    for source in sources:
        source_database = source.get("database")
        source_name = source["name"]
        source_identifier = source.get("schema", source_name)

        # Make folder for each source
        sub_folder = os.path.join(output_folder, source_name)
        os.makedirs(sub_folder, exist_ok=True)

        # Adapter to connect database
        adapter = get_adapter(profile, database=source_database)

        for table in source["tables"]:
            table_name = table["name"]
            table_identifier = table.get("identifier", table_name)

            # Identifier is name used in database
            columns = adapter.list_columns(source_identifier, table_identifier)

            # While, name is used in dbt
            sql_content = make_model_sql({"source_name": source_name, "table_name": table_name, "columns": columns})
            sql_path = os.path.join(sub_folder, f"stg__{table_name}.sql")
            write_file(sql_path, sql_content)


def generate_base_model(profile_path, output_folder, source_path, profile_name="default", target="dev", threads=None):
    os.makedirs(output_folder, exist_ok=True)

    profile = read_dbt_profile(profile_path, profile_name=profile_name, target=target)

    if os.path.isdir(source_path):
        with ThreadPoolExecutor(threads) as executor:
            futures = {}
            for fp in os.listdir(source_path):
                futures[
                    executor.submit(generate_base_from_sources, profile, output_folder, os.path.join(source_path, fp))
                ] = fp

            for ft in tqdm(as_completed(futures), total=len(futures)):
                pass
    elif os.path.isfile(source_path):
        generate_base_from_sources(profile, output_folder, source_path)
    else:
        raise ValueError(f"{source_path} is not a folder or file")

    print("Done")


def run(args):
    generate_base_model(
        profile_path=args.profile_path,
        output_folder=args.output_folder,
        source_path=args.source_path,
        profile_name=args.profile_name,
        target=args.target,
        threads=args.threads,
    )


def config_parser(parser):
    parser.set_defaults(func=run)
    parser.add_argument("source_path", type=str, help="Path to dbt source YAML.")
    parser.add_argument(
        "output_folder",
        type=str,
        help="Folder to write base models.",
    )
    parser.add_argument(
        "--profile-path",
        type=str,
        default=DEFAULT_DBT_PROFILE_PATH,
        help=("Path to dbt profile YAML." f" Default is {DEFAULT_DBT_PROFILE_PATH}"),
    )
    parser.add_argument("--profile-name", type=str, default="default", help="Dbt profile name. Default is `default`.")
    parser.add_argument("--target", type=str, default="dev", help="Dbt profile target. Default is `dev`.")
    parser.add_argument(
        "--threads",
        type=int,
        default=None,
        help="Max threads. Default is your machine number of threads.",
    )



from jinja2 import Environment, PackageLoader, select_autoescape, Template


class Templater:
    def __init__(self):
        self.env = Environment(loader=PackageLoader("dbt_gen"), autoescape=select_autoescape())

    def load_template(self, path: str) -> Template:
        return self.env.get_template(path)


_templater = None


def get_template(path: str) -> Template:
    global _templater
    if not _templater:
        _templater = Templater()
    return _templater.load_template(path)



"""Generate dbt source.yml file.
Read profiles.yml
Connect data warehouse
List all sources
Create source.yml file

Data warehouse supported:
- Snowflake
- BigQuery
"""
import os
from concurrent.futures import ThreadPoolExecutor, as_completed

from tqdm import tqdm

from .adapters import get_adapter
from .utils import DEFAULT_DBT_PROFILE_PATH, read_dbt_profile, write_file


def make_source_yaml(database, name, tables, schema=None):
    contents = [
        "version: 2",
        "",
        "sources:",
        f"  - name: {name}",
        f"    database: {database}",
    ]
    if schema:
        contents.append(f"    schema: {schema}")

    contents.append("    tables:")
    for tbl in tables:
        contents.append(f"      - name: {tbl.lower()}")

    contents.append("")
    return "\n".join(contents)


def generate_source(adapter, dataset, source_folder):
    dataset = dataset.lower()
    tables = adapter.list_tables(dataset)
    yaml_content = make_source_yaml(adapter.database, dataset, tables)
    yaml_path = os.path.join(source_folder, f"{dataset}.yml")
    write_file(yaml_path, yaml_content)


def generate_source_job(
    dbt_profile_path, source_folder, database=None, profile_name="default", target="dev", threads=None
):
    os.makedirs(source_folder, exist_ok=True)
    profile = read_dbt_profile(dbt_profile_path, profile_name=profile_name, target=target)
    adapter = get_adapter(profile, database=database)

    with ThreadPoolExecutor(threads) as executor:
        futures = {}
        for dataset in adapter.list_datasets():
            futures[executor.submit(generate_source, adapter, dataset, source_folder)] = dataset

        for ft in tqdm(as_completed(futures), total=len(futures)):
            pass

    print("Done")


def run(args):
    generate_source_job(
        dbt_profile_path=args.profile_path,
        source_folder=args.source_folder,
        profile_name=args.profile_name,
        target=args.target,
        database=args.database,
        threads=args.threads,
    )


def config_parser(parser):
    parser.set_defaults(func=run)
    parser.add_argument(
        "source_folder",
        type=str,
        help="Folder to write source YAML files",
    )
    parser.add_argument(
        "--profile-path",
        type=str,
        default=DEFAULT_DBT_PROFILE_PATH,
        help=("Path to dbt profile YAML." f" Default is {DEFAULT_DBT_PROFILE_PATH}"),
    )
    parser.add_argument("--profile-name", type=str, default="default", help="Dbt profile name. Default is `default`.")
    parser.add_argument("--target", type=str, default="dev", help="Dbt profile target. Default is `dev`.")
    parser.add_argument(
        "--database",
        type=str,
        default=None,
        help="Database to inspect. Default is the database in profile.",
    )
    parser.add_argument(
        "--threads",
        type=int,
        default=None,
        help="Max threads. Default is your machine number of threads.",
    )



from argparse import ArgumentParser

from . import generate_source_yaml
from . import generate_base_model
from . import generate_dbt


def add_parser(sub_parsers, name, config_func):
    subparser = sub_parsers.add_parser(name)
    config_func(subparser)
    return subparser


def cli():
    parser = ArgumentParser()
    sub_parsers = parser.add_subparsers(help="Action")

    add_parser(sub_parsers, "generate_source", generate_source_yaml.config_parser)
    add_parser(sub_parsers, "generate_base_model", generate_base_model.config_parser)
    add_parser(sub_parsers, "generate_dbt", generate_dbt.config_parser)

    args = parser.parse_args()

    if "func" in args:
        args.func(args)
    else:
        print("Please choose action\n")
        parser.print_help()


def main():
    cli()



import os
from pathlib import Path

import yaml

DEFAULT_DBT_PROFILE_PATH = os.path.join(str(Path.home()), ".dbt", "profiles.yml")


def read_yaml(path):
    with open(path, "r", encoding="utf-8") as file:
        return yaml.load(file, Loader=yaml.FullLoader)


def read_dbt_profile(file_path, profile_name="default", target="dev"):
    profiles = read_yaml(file_path)

    if profile_name not in profiles:
        raise ValueError(f"Profile `{profile_name}` not found.")

    profile = profiles[profile_name]

    if target not in profile["outputs"]:
        raise ValueError(f"Target `{target}` not found.")

    return profile["outputs"][target]


def write_file(file_path, text):
    with open(file_path, "w", encoding="utf-8") as file:
        file.write(text)



from .base import BaseAdapter
from .bigquery_adapter import BigQueryAdapter
from .snowflake_adapter import SnowflakeAdapter


def get_adapter(profile: dict, database=None) -> BaseAdapter:
    adapter_name = profile["type"]
    if adapter_name == "bigquery":
        return BigQueryAdapter.from_dbt_profile(profile, database=database)
    elif adapter_name == "snowflake":
        return SnowflakeAdapter.from_dbt_profile(profile, database=database)
    else:
        raise ValueError(f"Adapter {adapter_name} not exists")



import pathlib

from setuptools import find_packages, setup

MAIN_REQUIREMENTS = [
    "pyarrow==6.0.1",
    "pyYAML==6.0",
    "pandas==1.3.4",
    "google-cloud-bigquery==2.34.0",
    "snowflake-connector-python==2.7.4",
    "jinja2==3.0.3",
    "tqdm==4.63.0",
]

TEST_REQUIREMENTS = [
    "flake8==4.0.1",
    "black==22.1.0",
    "pytest==7.0.1",
]

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="dbt_gen",
    version="0.1.3",
    long_description=README,
    long_description_content_type="text/markdown",
    description="Tool to generate dbt resources.",
    url="https://github.com/joon-solutions/dbt-gen",
    author="Joon Solutions",
    author_email="tien.tq@joonsolutions.com",
    license="GPLv3",
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "dbt-gen=dbt_gen.cli:main"
        ]
    },
    install_requires=MAIN_REQUIREMENTS,
    package_data={"": ["templates/*.sql"]},
    extras_require={
        "tests": TEST_REQUIREMENTS,
    },
)



import logging
from functools import lru_cache

import pandas as pd
import snowflake.connector
from .base import ColumnInfo, BaseAdapter

logger = logging.getLogger(__name__)


class SnowflakeAdapter(BaseAdapter):
    def __init__(self, user, password, account, warehouse, role, database):
        self.user = user
        self._password = password
        self.account = account
        self.warehouse = warehouse
        self.role = role
        self.database = database
        self._conn = None

    @classmethod
    def from_dbt_profile(cls, profile: dict, database=None):
        database = database or profile["database"]
        return cls(
            user=profile["user"],
            password=profile["password"],
            account=profile["account"],
            warehouse=profile["warehouse"],
            role=profile["role"],
            database=database,
        )

    @property
    def _connection(self):
        if not self._conn:
            self._conn = snowflake.connector.connect(
                user=self.user,
                password=self._password,
                account=self.account,
                warehouse=self.warehouse,
                role=self.role,
                database=self.database,
            )
        return self._conn

    def execute(self, query, **kwargs):
        cursor = self._connection.cursor()
        try:
            cursor.execute(query, **kwargs)
            return cursor.fetchall()
        finally:
            cursor.close()

    def execute_pandas(self, query, **kwargs) -> pd.DataFrame:
        cursor = self._connection.cursor()
        try:
            cursor.execute(query, **kwargs)
            return cursor.fetch_pandas_all()
        finally:
            cursor.close()

    @lru_cache()
    def get_column_metadata_df(self, dataset) -> pd.DataFrame:
        query = f"select * from information_schema.columns where table_schema = '{dataset.upper()}';"
        dataframe = self.execute_pandas(query)
        dataframe.columns = dataframe.columns.str.lower()
        dataframe = dataframe.loc[
            dataframe["table_schema"].apply(lambda x: (x.lower() != "information_schema" and not x.startswith("_")))
            & dataframe["table_name"].apply(lambda x: not x.startswith("_")),
            :,
        ]
        dataframe.sort_values(by=["table_schema", "table_name", "ordinal_position"], inplace=True)
        return dataframe

    def list_datasets(self):
        query = "select distinct schema_name from information_schema.schemata;"
        meta_df = self.execute_pandas(query)
        for schema in meta_df["SCHEMA_NAME"].unique():
            yield schema.lower()

    def list_tables(self, dataset):
        meta_df = self.get_column_metadata_df(dataset)
        meta_df = meta_df.loc[meta_df["table_schema"] == dataset.upper(), :]
        for table in meta_df["table_name"].unique():
            yield table.lower()

    def list_columns(self, dataset, table):
        meta_df = self.get_column_metadata_df(dataset)
        meta_df = meta_df.loc[
            (meta_df["table_schema"] == dataset.upper()) & (meta_df["table_name"] == table.upper()), :
        ]
        rows = meta_df.to_dict(orient="records")
        for row in rows:
            yield ColumnInfo(
                database=self.database,
                schema=dataset,
                table=table,
                name=row["column_name"].lower(),
                position=row["ordinal_position"],
                dtype=row["data_type"],
                nullable=True if row["is_nullable"] == "YES" else False,
                extra=row,
            )

    # def __del__(self):
    #     if self._conn:
    #         print("Closing Snowflake connection")
    #         self._conn.close()



from cookiecutter.main import cookiecutter

DEFAULT_GIT_URL = "https://github.com/joon-solutions/cookiecutter-dbt.git"


def generate_dbt(**_kwarg):
    print(f"Enter Git URL [{DEFAULT_GIT_URL}]: ")
    git_url = input()
    if not git_url:
        git_url = DEFAULT_GIT_URL
    cookiecutter(git_url)


def run(args):
    generate_dbt(**(args.__dict__))


def config_parser(parser):
    parser.set_defaults(func=run)
